from __future__ import annotations

import pytest

from pyjelly import jelly
from pyjelly.errors import JellyAssertionError
from pyjelly.options import StreamTypes


@pytest.mark.parametrize(
    ("physical_type", "logical_type"),
    VALID_STREAM_TYPE_COMBINATIONS := [
        (jelly.PHYSICAL_STREAM_TYPE_TRIPLES, jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES),
        (jelly.PHYSICAL_STREAM_TYPE_TRIPLES, jelly.LOGICAL_STREAM_TYPE_GRAPHS),
        (jelly.PHYSICAL_STREAM_TYPE_TRIPLES, jelly.LOGICAL_STREAM_TYPE_SUBJECT_GRAPHS),
        (jelly.PHYSICAL_STREAM_TYPE_TRIPLES, jelly.LOGICAL_STREAM_TYPE_UNSPECIFIED),
        (jelly.PHYSICAL_STREAM_TYPE_QUADS, jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS),
        (jelly.PHYSICAL_STREAM_TYPE_QUADS, jelly.LOGICAL_STREAM_TYPE_DATASETS),
        (jelly.PHYSICAL_STREAM_TYPE_QUADS, jelly.LOGICAL_STREAM_TYPE_NAMED_GRAPHS),
        (
            jelly.PHYSICAL_STREAM_TYPE_QUADS,
            jelly.LOGICAL_STREAM_TYPE_TIMESTAMPED_NAMED_GRAPHS,
        ),
        (jelly.PHYSICAL_STREAM_TYPE_QUADS, jelly.LOGICAL_STREAM_TYPE_UNSPECIFIED),
        (jelly.PHYSICAL_STREAM_TYPE_GRAPHS, jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS),
        (jelly.PHYSICAL_STREAM_TYPE_GRAPHS, jelly.LOGICAL_STREAM_TYPE_DATASETS),
        (jelly.PHYSICAL_STREAM_TYPE_GRAPHS, jelly.LOGICAL_STREAM_TYPE_NAMED_GRAPHS),
        (
            jelly.PHYSICAL_STREAM_TYPE_GRAPHS,
            jelly.LOGICAL_STREAM_TYPE_TIMESTAMPED_NAMED_GRAPHS,
        ),
        (jelly.PHYSICAL_STREAM_TYPE_GRAPHS, jelly.LOGICAL_STREAM_TYPE_UNSPECIFIED),
        (jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED, jelly.LOGICAL_STREAM_TYPE_UNSPECIFIED),
        (
            jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED,
            jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES,
        ),
        (jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED, jelly.LOGICAL_STREAM_TYPE_GRAPHS),
        (
            jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED,
            jelly.LOGICAL_STREAM_TYPE_SUBJECT_GRAPHS,
        ),
        (jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED, jelly.LOGICAL_STREAM_TYPE_UNSPECIFIED),
        (jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED, jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS),
        (jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED, jelly.LOGICAL_STREAM_TYPE_DATASETS),
        (
            jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED,
            jelly.LOGICAL_STREAM_TYPE_NAMED_GRAPHS,
        ),
        (
            jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED,
            jelly.LOGICAL_STREAM_TYPE_TIMESTAMPED_NAMED_GRAPHS,
        ),
    ],
)
def test_stream_types_ok(
    physical_type: jelly.PhysicalStreamType,
    logical_type: jelly.LogicalStreamType,
) -> None:
    StreamTypes(physical_type=physical_type, logical_type=logical_type)


@pytest.mark.parametrize(
    ("physical_type", "logical_type"),
    [
        (jelly.PHYSICAL_STREAM_TYPE_TRIPLES, jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS),
        (jelly.PHYSICAL_STREAM_TYPE_TRIPLES, jelly.LOGICAL_STREAM_TYPE_DATASETS),
        (jelly.PHYSICAL_STREAM_TYPE_TRIPLES, jelly.LOGICAL_STREAM_TYPE_NAMED_GRAPHS),
        (
            jelly.PHYSICAL_STREAM_TYPE_TRIPLES,
            jelly.LOGICAL_STREAM_TYPE_TIMESTAMPED_NAMED_GRAPHS,
        ),
        (jelly.PHYSICAL_STREAM_TYPE_QUADS, jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES),
        (jelly.PHYSICAL_STREAM_TYPE_QUADS, jelly.LOGICAL_STREAM_TYPE_GRAPHS),
        (jelly.PHYSICAL_STREAM_TYPE_QUADS, jelly.LOGICAL_STREAM_TYPE_SUBJECT_GRAPHS),
        (jelly.PHYSICAL_STREAM_TYPE_GRAPHS, jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES),
        (jelly.PHYSICAL_STREAM_TYPE_GRAPHS, jelly.LOGICAL_STREAM_TYPE_GRAPHS),
        (jelly.PHYSICAL_STREAM_TYPE_GRAPHS, jelly.LOGICAL_STREAM_TYPE_SUBJECT_GRAPHS),
    ],
)
def test_stream_types_error(
    physical_type: jelly.PhysicalStreamType,
    logical_type: jelly.LogicalStreamType,
) -> None:
    with pytest.raises(JellyAssertionError):
        StreamTypes(physical_type=physical_type, logical_type=logical_type)



from rdflib import Dataset, Graph

# libraries to load example jelly stream data
import gzip
import urllib.request
from typing import cast, IO

from pyjelly.integrations.rdflib.parse import parse_jelly_grouped

url = "https://w3id.org/riverbench/datasets/dbpedia-live/dev/files/jelly_10K.jelly.gz"

# load, uncompress .gz file, and pass to jelly parser
with (
    urllib.request.urlopen(url) as resp,
    cast(IO[bytes], gzip.GzipFile(fileobj=resp)) as jelly_stream,
):
    graphs = parse_jelly_grouped(
        jelly_stream,
        graph_factory=lambda: Graph(),
        dataset_factory=lambda: Dataset(),
    )
    for i, graph in enumerate(graphs):
        print(f"Graph {i} in the stream has {len(graph)} triples")



from __future__ import annotations

from pathlib import Path

import pytest

from tests.meta import (
    RDF_TO_JELLY_TESTS_DIR,
    TEST_OUTPUTS_DIR,
)
from tests.serialize import write_graph_or_dataset
from tests.utils.rdf_test_cases import (
    PhysicalTypeTestCasesDir,
    id_from_path,
    jelly_validate,
    needs_jelly_cli,
    walk_directories,
)


@needs_jelly_cli
@walk_directories(
    RDF_TO_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.TRIPLES,
    RDF_TO_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.QUADS,
    RDF_TO_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.GRAPHS,
    glob="pos_*",
)
def test_serializes(path: Path) -> None:
    options_filename = path / "stream_options.jelly"
    input_filenames = tuple(path.glob("in_*"))
    test_id = id_from_path(path)
    actual_out = TEST_OUTPUTS_DIR / f"{test_id}.jelly"

    write_graph_or_dataset(
        *input_filenames,
        options=options_filename,
        out_filename=actual_out,
    )
    for input_filename in input_filenames:
        jelly_validate(
            actual_out,
            "--compare-ordered",
            "--compare-to-rdf-file",
            input_filename,
            "--options-file",
            options_filename,
            hint=f"Test ID: {test_id}, tested file: {input_filename}",
        )


@needs_jelly_cli
@walk_directories(
    RDF_TO_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.TRIPLES,
    RDF_TO_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.QUADS,
    RDF_TO_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.GRAPHS,
    glob="neg_*",
)
def test_serializing_fails(path: Path) -> None:
    options_filename = path / "stream_options.jelly"
    test_id = id_from_path(path)
    actual_out = TEST_OUTPUTS_DIR / f"{test_id}.jelly"
    with pytest.raises(Exception):  # TODO: more specific  # noqa: PT011,B017,TD002
        write_graph_or_dataset(
            *map(str, path.glob("in_*")),
            options=options_filename,
            out_filename=actual_out,
        )



import pytest
from hypothesis import given
from hypothesis import strategies as st

from pyjelly.errors import JellyAssertionError
from pyjelly.options import MAX_LOOKUP_SIZE
from pyjelly.parse.lookup import LookupDecoder


@given(st.integers(min_value=1, max_value=MAX_LOOKUP_SIZE))
def test_lookup_size_ok(size: int) -> None:
    LookupDecoder(lookup_size=size)


@given(st.integers(min_value=MAX_LOOKUP_SIZE + 1))
def test_max_lookup_size_exceeded(size: int) -> None:
    with pytest.raises(JellyAssertionError) as excinfo:
        LookupDecoder(lookup_size=size)
    assert str(excinfo.value) == f"lookup size must be less than {MAX_LOOKUP_SIZE}"



from __future__ import annotations

from collections.abc import Generator, Iterable
from typing import IO, Any, Callable
from typing_extensions import Never, override

import rdflib
from rdflib.graph import DATASET_DEFAULT_GRAPH_ID, Dataset, Graph
from rdflib.parser import InputSource
from rdflib.parser import Parser as RDFLibParser

from pyjelly import jelly
from pyjelly.errors import JellyConformanceError
from pyjelly.options import StreamTypes
from pyjelly.parse.decode import Adapter, Decoder, ParserOptions, ParsingMode
from pyjelly.parse.ioutils import get_options_and_frames


class RDFLibAdapter(Adapter):
    """
    RDFLib adapter class, is extended by triples and quads implementations.

    Args:
        Adapter (_type_): abstract adapter class

    """

    @override
    def iri(self, iri: str) -> rdflib.URIRef:
        return rdflib.URIRef(iri)

    @override
    def bnode(self, bnode: str) -> rdflib.BNode:
        return rdflib.BNode(bnode)

    @override
    def default_graph(self) -> rdflib.URIRef:
        return DATASET_DEFAULT_GRAPH_ID

    @override
    def literal(
        self,
        lex: str,
        language: str | None = None,
        datatype: str | None = None,
    ) -> rdflib.Literal:
        return rdflib.Literal(lex, lang=language, datatype=datatype)


def _adapter_missing(feature: str, *, stream_types: StreamTypes) -> Never:
    """
    Raise error if functionality is missing in adapter.

    TODO: currently not used anywhere due to logical types being removed

    Args:
        feature (str): function which is not implemented
        stream_types (StreamTypes): what combination of physical/logical types
            triggered the error

    Raises:
        NotImplementedError: raises error with message with missing functionality
            and types encountered

    Returns:
        Never: only raises errors

    """
    physical_type_name = jelly.PhysicalStreamType.Name(stream_types.physical_type)
    logical_type_name = jelly.LogicalStreamType.Name(stream_types.logical_type)
    msg = (
        f"adapter with {physical_type_name} and {logical_type_name} "
        f"does not implement {feature}"
    )
    raise NotImplementedError(msg)


class RDFLibTriplesAdapter(RDFLibAdapter):
    """
    Triples adapter RDFLib implementation.

    Notes: has internal graph object which tracks
        triples and namespaces and can get flushed between frames.
    """

    def __init__(
        self,
        options: ParserOptions,
        graph_factory: Callable[[], Graph],
        parsing_mode: ParsingMode = ParsingMode.FLAT,
    ) -> None:
        super().__init__(options=options, parsing_mode=parsing_mode)
        self.graph = graph_factory()
        self.graph_factory = graph_factory
        self.parsing_mode = parsing_mode

    @override
    def triple(self, terms: Iterable[Any]) -> Any:
        self.graph.add(tuple(terms))

    @override
    def namespace_declaration(self, name: str, iri: str) -> None:
        self.graph.bind(name, self.iri(iri))

    def frame(self) -> Graph:
        """
        Finalize one frame in triples stream.

        Returns:
           Graph: frame content as a separate Graph
                and starts a new Graph

        """
        this_graph = self.graph
        self.graph = self.graph_factory()
        return this_graph


class RDFLibQuadsBaseAdapter(RDFLibAdapter):
    def __init__(
        self,
        options: ParserOptions,
        dataset_factory: Callable[[], Dataset],
        parsing_mode: ParsingMode = ParsingMode.FLAT,
    ) -> None:
        super().__init__(options=options, parsing_mode=parsing_mode)
        self.dataset = dataset_factory()
        self.dataset_factory = dataset_factory

    @override
    def frame(self) -> Dataset:
        current_dataset = self.dataset
        self.dataset = self.dataset_factory()
        return current_dataset


class RDFLibQuadsAdapter(RDFLibQuadsBaseAdapter):
    """
    Extended RDFLib adapter for the QUADS physical type.

    Notes:
        Adds triples and namespaces directly to
        dataset, so RDFLib handles the rest.

    Args:
        RDFLibQuadsBaseAdapter (_type_): base quads adapter
            (shared with graphs physical type)

    """

    @override
    def namespace_declaration(self, name: str, iri: str) -> None:
        self.dataset.bind(name, self.iri(iri))

    @override
    def quad(self, terms: Iterable[Any]) -> Any:
        self.dataset.add(tuple(terms))


class RDFLibGraphsAdapter(RDFLibQuadsBaseAdapter):
    """
    Extension of RDFLibQuadsBaseAdapter for the GRAPHS physical type.

    Notes: introduces graph start/end, checks if graph exists,
        dataset store management.

    Args:
        RDFLibQuadsBaseAdapter (_type_): base adapter for quads management.

    Raises:
        JellyConformanceError: if no graph_start was encountered

    """

    _graph_id: str | None

    def __init__(
        self,
        options: ParserOptions,
        dataset_factory: Callable[[], Dataset],
        parsing_mode: ParsingMode = ParsingMode.FLAT,
    ) -> None:
        super().__init__(
            options=options,
            dataset_factory=dataset_factory,
            parsing_mode=parsing_mode,
        )
        self._graph_id = None

    @property
    def graph(self) -> None:
        if self._graph_id is None:
            msg = "new graph was not started"
            raise JellyConformanceError(msg)

    @override
    def graph_start(self, graph_id: str) -> None:
        self._graph_id = graph_id

    @override
    def namespace_declaration(self, name: str, iri: str) -> None:
        self.dataset.bind(name, self.iri(iri))

    @override
    def triple(self, terms: Iterable[Any]) -> None:
        self.dataset.add((*terms, self._graph_id))

    @override
    def graph_end(self) -> None:
        self._graph_id = None


def parse_triples_stream(
    frames: Iterable[jelly.RdfStreamFrame],
    options: ParserOptions,
    graph_factory: Callable[[], Graph],
    parsing_mode: ParsingMode = ParsingMode.FLAT,
) -> Generator[Graph]:
    """
    Parse flat triple stream.

    Args:
        frames (Iterable[jelly.RdfStreamFrame]): iterator over stream frames
        options (ParserOptions): stream options
        graph_factory (Callable): Lambda to construct a graph
        parsing_mode (ParsingMode): specifies whether this is
            a flat or grouped parsing.

    Yields:
        Generator[Graph]: RDFLib Graph(s)

    """
    adapter = RDFLibTriplesAdapter(
        options, graph_factory=graph_factory, parsing_mode=parsing_mode
    )
    decoder = Decoder(adapter=adapter)
    for frame in frames:
        g = decoder.decode_frame(frame)
        if g is not None:
            yield g

    if parsing_mode is ParsingMode.FLAT:
        yield adapter.graph


def parse_quads_stream(
    frames: Iterable[jelly.RdfStreamFrame],
    options: ParserOptions,
    dataset_factory: Callable[[], Dataset],
    parsing_mode: ParsingMode = ParsingMode.FLAT,
) -> Generator[Dataset]:
    """
    Parse flat quads stream.

    Args:
        frames (Iterable[jelly.RdfStreamFrame]): iterator over stream frames
        options (ParserOptions): stream options
        dataset_factory (Callable): Lambda to construct a dataset
        parsing_mode (ParsingMode): specifies whether this is
            a flat or grouped parsing.

    Yields:
        Generator[Dataset]: RDFLib dataset(s)

    """
    adapter_class: type[RDFLibQuadsBaseAdapter]
    if options.stream_types.physical_type == jelly.PHYSICAL_STREAM_TYPE_QUADS:
        adapter_class = RDFLibQuadsAdapter
    else:
        adapter_class = RDFLibGraphsAdapter
    adapter = adapter_class(
        options=options,
        dataset_factory=dataset_factory,
        parsing_mode=parsing_mode,
    )
    decoder = Decoder(adapter=adapter)
    for frame in frames:
        ds = decoder.decode_frame(frame)
        if ds is not None:
            yield ds

    if parsing_mode is ParsingMode.FLAT:
        yield adapter.dataset


def parse_jelly_grouped(
    inp: IO[bytes],
    graph_factory: Callable[[], Graph],
    dataset_factory: Callable[[], Dataset],
) -> Generator[Any] | Generator[Graph] | Generator[Dataset]:
    """
    Take jelly file and return generators based on the detected logical type.

    Yields one graph/dataset per frame.

    Args:
        inp (IO[bytes]): input jelly buffered binary stream
        graph_factory (Callable): lambda to construct a Graph
        dataset_factory (Callable): lambda to construct a Dataset

    Raises:
        NotImplementedError: is raised if a logical type is not implemented

    Yields:
        Generator[Any] | Generator[Dataset] | Generator[Graph]:
            returns generators for graphs/datasets based on the type of input

    """
    options, frames = get_options_and_frames(inp)

    if options.stream_types.physical_type == jelly.PHYSICAL_STREAM_TYPE_TRIPLES:
        yield from parse_triples_stream(
            frames=frames,
            options=options,
            graph_factory=graph_factory,
            parsing_mode=ParsingMode.GROUPED,
        )
        return

    if options.stream_types.physical_type in (
        jelly.PHYSICAL_STREAM_TYPE_QUADS,
        jelly.PHYSICAL_STREAM_TYPE_GRAPHS,
    ):
        yield from parse_quads_stream(
            frames=frames,
            options=options,
            dataset_factory=dataset_factory,
            parsing_mode=ParsingMode.GROUPED,
        )
        return

    physical_type_name = jelly.PhysicalStreamType.Name(
        options.stream_types.physical_type
    )
    msg = f"the stream type {physical_type_name} is not supported "
    raise NotImplementedError(msg)


def parse_jelly_flat(
    inp: IO[bytes],
    graph_factory: Callable[[], Graph],
    dataset_factory: Callable[[], Dataset],
) -> Any | Dataset | Graph:
    """
    Parse jelly file with FLAT physical type into one Graph/Dataset.

    Args:
        inp (IO[bytes]): input jelly buffered binary stream
        graph_factory (Callable): lambda to construct a Graph
        dataset_factory (Callable): lambda to construct a Dataset

    Raises:
        NotImplementedError: if physical type is not supported

    Returns:
        RDFLib Graph or Dataset

    """
    options, frames = get_options_and_frames(inp)

    if options.stream_types.physical_type == jelly.PHYSICAL_STREAM_TYPE_TRIPLES:
        return next(
            parse_triples_stream(
                frames=frames,
                options=options,
                graph_factory=graph_factory,
                parsing_mode=ParsingMode.FLAT,
            )
        )

    if options.stream_types.physical_type in (
        jelly.PHYSICAL_STREAM_TYPE_QUADS,
        jelly.PHYSICAL_STREAM_TYPE_GRAPHS,
    ):
        return next(
            parse_quads_stream(
                frames=frames,
                options=options,
                dataset_factory=dataset_factory,
                parsing_mode=ParsingMode.FLAT,
            )
        )
    physical_type_name = jelly.PhysicalStreamType.Name(
        options.stream_types.physical_type
    )
    msg = f"the stream type {physical_type_name} is not supported "
    raise NotImplementedError(msg)


class RDFLibJellyParser(RDFLibParser):
    def parse(self, source: InputSource, sink: Graph) -> None:
        """
        Parse jelly file into provided RDFLib Graph.

        Args:
            source (InputSource): jelly file as buffered binary stream InputSource obj
            sink (Graph): RDFLib Graph

        Raises:
            TypeError: raises error if invalid input

        """
        inp = source.getByteStream()  # type: ignore[no-untyped-call]
        if inp is None:
            msg = "expected source to be a stream of bytes"
            raise TypeError(msg)
        parse_jelly_flat(
            inp,
            graph_factory=lambda: Graph(store=sink.store, identifier=sink.identifier),
            dataset_factory=lambda: Dataset(store=sink.store),
        )



from __future__ import annotations

from collections.abc import Generator
from functools import singledispatch
from typing import IO, Any
from typing_extensions import override

import rdflib
from rdflib.graph import DATASET_DEFAULT_GRAPH_ID, Dataset, Graph, QuotedGraph
from rdflib.serializer import Serializer as RDFLibSerializer

from pyjelly import jelly
from pyjelly.serialize.encode import RowsAndTerm, Slot, TermEncoder
from pyjelly.serialize.ioutils import write_delimited, write_single
from pyjelly.serialize.streams import (
    GraphStream,
    QuadStream,
    SerializerOptions,
    Stream,
    TripleStream,
)


class RDFLibTermEncoder(TermEncoder):
    def encode_any(self, term: object, slot: Slot) -> RowsAndTerm:
        """
        Encode term based on its RDFLib object.

        Args:
            term (object): term to encode
            slot (Slot): its place in statement.

        Returns:
            RowsAndTerm: encoded extra rows and a jelly term to encode

        """
        if slot is Slot.graph and term == DATASET_DEFAULT_GRAPH_ID:
            return self.encode_default_graph()

        if isinstance(term, rdflib.URIRef):
            return self.encode_iri(term)

        if isinstance(term, rdflib.Literal):
            return self.encode_literal(
                lex=str(term),
                language=term.language,
                # `datatype` is cast to `str` explicitly because
                # `URIRef.__eq__` overrides `str.__eq__` in an incompatible manner
                datatype=term.datatype and str(term.datatype),
            )

        if isinstance(term, rdflib.BNode):
            return self.encode_bnode(str(term))

        return super().encode_any(term, slot)  # error if not handled


def namespace_declarations(store: Graph, stream: Stream) -> None:
    for prefix, namespace in store.namespaces():
        stream.namespace_declaration(name=prefix, iri=namespace)


@singledispatch
def stream_frames(stream: Stream, data: Graph) -> Generator[jelly.RdfStreamFrame]:  # noqa: ARG001
    msg = f"invalid stream implementation {stream}"
    raise TypeError(msg)


@stream_frames.register(TripleStream)
def triples_stream_frames(
    stream: TripleStream,
    data: Graph | Dataset,
) -> Generator[jelly.RdfStreamFrame]:
    """
    Serialize a Graph/Dataset into jelly frames.

    Args:
        stream (TripleStream): stream that specifies triples processing
        data (Graph | Dataset): Graph/Dataset to serialize.

    Notes:
        if Dataset is given, its graphs are unpacked and iterated over
        if flow is GraphsFrameFlow, emits a frame per graph.

    Yields:
        Generator[jelly.RdfStreamFrame]: jelly frames.

    """
    stream.enroll()
    if stream.options.params.namespace_declarations:
        namespace_declarations(data, stream)
    graphs = (data,) if not isinstance(data, Dataset) else data.graphs()
    for graph in graphs:
        for terms in graph:
            if frame := stream.triple(terms):
                yield frame
        # this part turns each graph to a frame for graphs logical type
        if frame := stream.flow.frame_from_graph():
            yield frame
    if stream.stream_types.flat and (frame := stream.flow.to_stream_frame()):
        yield frame


@stream_frames.register
def quads_stream_frames(
    stream: QuadStream,
    data: Dataset,
) -> Generator[jelly.RdfStreamFrame]:
    """
    Serialize a Dataset into jelly frames.

    Notes:
        Emits one frame per dataset if flow is of DatasetsFrameFlow.

    Args:
        stream (QuadStream): stream that specifies quads processing
        data (Dataset): Dataset to serialize.

    Yields:
        Generator[jelly.RdfStreamFrame]: jelly frames

    """
    assert isinstance(data, Dataset)
    stream.enroll()
    if stream.options.params.namespace_declarations:
        namespace_declarations(data, stream)
    for terms in data.quads():
        if frame := stream.quad(terms):
            yield frame
    if frame := stream.flow.frame_from_dataset():
        yield frame
    if stream.stream_types.flat and (frame := stream.flow.to_stream_frame()):
        yield frame


@stream_frames.register
def graphs_stream_frames(
    stream: GraphStream,
    data: Dataset,
) -> Generator[jelly.RdfStreamFrame]:
    """
    Serialize a Dataset into jelly frames as a stream of graphs.

    Notes:
        If flow of DatasetsFrameFlow type, the whole dataset
        will be encoded into one frame.

    Args:
        stream (GraphStream): stream that specifies graphs processing
        data (Dataset): Dataset to serialize.

    Yields:
        Generator[jelly.RdfStreamFrame]: jelly frames

    """
    assert isinstance(data, Dataset)
    stream.enroll()
    if stream.options.params.namespace_declarations:
        namespace_declarations(data, stream)
    for graph in data.graphs():
        yield from stream.graph(graph_id=graph.identifier, graph=graph)
    if frame := stream.flow.frame_from_dataset():
        yield frame
    if stream.stream_types.flat and (frame := stream.flow.to_stream_frame()):
        yield frame


class RDFLibJellySerializer(RDFLibSerializer):
    """
    RDFLib serializer for writing graphs in Jelly RDF stream format.

    Handles streaming RDF terms into Jelly frames using internal encoders.
    Supports only graphs and datasets (not quoted graphs).

    """

    def __init__(self, store: Graph) -> None:
        if isinstance(store, QuotedGraph):
            msg = "N3 format is not supported"
            raise NotImplementedError(msg)
        super().__init__(store)

    def guess_options(self) -> SerializerOptions:
        """
        Guess the serializer options based on the store type.

        >>> RDFLibJellySerializer(Graph()).guess_options().logical_type
        1
        >>> RDFLibJellySerializer(Dataset()).guess_options().logical_type
        2
        """
        logical_type = (
            jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS
            if isinstance(self.store, Dataset)
            else jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES
        )
        return SerializerOptions(logical_type=logical_type)

    def guess_stream(self, options: SerializerOptions) -> Stream:
        """
        Return an appropriate stream implementation for the given options.

        >>> graph_ser = RDFLibJellySerializer(Graph())
        >>> ds_ser = RDFLibJellySerializer(Dataset())

        >>> type(graph_ser.guess_stream(graph_ser.guess_options()))
        <class 'pyjelly.serialize.streams.TripleStream'>
        >>> type(ds_ser.guess_stream(ds_ser.guess_options()))
        <class 'pyjelly.serialize.streams.QuadStream'>
        """
        stream_cls: type[Stream]
        if options.logical_type != jelly.LOGICAL_STREAM_TYPE_GRAPHS and isinstance(
            self.store, Dataset
        ):
            stream_cls = QuadStream
        else:
            stream_cls = TripleStream
        return stream_cls.for_rdflib(options=options)

    @override
    def serialize(  # type: ignore[override]
        self,
        out: IO[bytes],
        /,
        *,
        stream: Stream | None = None,
        options: SerializerOptions | None = None,
        **unused: Any,
    ) -> None:
        """
        Serialize self.store content to Jelly format.

        Args:
            out (IO[bytes]): output buffered writer
            stream (Stream | None, optional): Jelly stream object. Defaults to None.
            options (SerializerOptions | None, optional): Serializer options
                if defined beforehand, e.g., read from a separate file.
                Defaults to None.
            **unused(Any): unused args for RDFLib serialize

        """
        if options is None:
            options = self.guess_options()
        if stream is None:
            stream = self.guess_stream(options)
        write = write_delimited if stream.options.params.delimited else write_single
        for stream_frame in stream_frames(stream, self.store):
            write(stream_frame, out)



from __future__ import annotations

from collections import UserList
from collections.abc import Iterable
from dataclasses import dataclass
from typing import ClassVar
from typing_extensions import override

from pyjelly import jelly

DEFAULT_FRAME_SIZE = 250


class FrameFlow(UserList[jelly.RdfStreamRow]):
    """
    Abstract base class for producing Jelly frames from RDF stream rows.

    Collects stream rows and assembles them into RdfStreamFrame objects when ready.
    """

    logical_type: jelly.LogicalStreamType
    registry: ClassVar[dict[jelly.LogicalStreamType, type[FrameFlow]]] = {}

    def frame_from_graph(self) -> jelly.RdfStreamFrame | None:
        """
        Treat the current rows as a graph and produce a frame.

        Default implementation returns None.
        """
        return None

    def frame_from_dataset(self) -> jelly.RdfStreamFrame | None:
        """
        Treat the current rows as a dataset and produce a frame.

        Default implementation returns None.
        """
        return None

    def frame_from_bounds(self) -> jelly.RdfStreamFrame | None:
        return None

    def to_stream_frame(self) -> jelly.RdfStreamFrame | None:
        """
        Create stream frame from flow content.

        Notes:
            Clears flow content after creating the frame.

        Returns:
            jelly.RdfStreamFrame | None: stream frame

        """
        if not self:
            return None
        frame = jelly.RdfStreamFrame(rows=self)
        self.clear()
        return frame


class ManualFrameFlow(FrameFlow):
    """
    Produces frames only when manually requested (never automatically).

    !!! warning
        All stream rows are kept in memory until `to_stream_frame()` is called.
        This may lead to high memory usage for large streams.

    Used for non-delimited serialization.
    """

    logical_type = jelly.LOGICAL_STREAM_TYPE_UNSPECIFIED

    def __init__(
        self,
        initlist: Iterable[jelly.RdfStreamRow] | None = None,
        *,
        logical_type: jelly.LogicalStreamType = jelly.LOGICAL_STREAM_TYPE_UNSPECIFIED,
    ) -> None:
        super().__init__(initlist)
        self.logical_type = logical_type


@dataclass
class BoundedFrameFlow(FrameFlow):
    """
    Produce frames automatically when a fixed number of rows is reached.

    Used for delimited encoding (default mode).
    """

    logical_type = jelly.LOGICAL_STREAM_TYPE_UNSPECIFIED
    frame_size: int

    def __init__(
        self,
        initlist: Iterable[jelly.RdfStreamRow] | None = None,
        *,
        frame_size: int | None = None,
    ) -> None:
        super().__init__(initlist)
        self.frame_size = frame_size or DEFAULT_FRAME_SIZE

    @override
    def frame_from_bounds(self) -> jelly.RdfStreamFrame | None:
        """
        Emit frame from flow if full.

        Returns:
            jelly.RdfStreamFrame | None: stream frame

        """
        if len(self) >= self.frame_size:
            return self.to_stream_frame()
        return None


class FlatTriplesFrameFlow(BoundedFrameFlow):
    logical_type = jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES


class FlatQuadsFrameFlow(BoundedFrameFlow):
    logical_type = jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS


class GraphsFrameFlow(FrameFlow):
    logical_type = jelly.LOGICAL_STREAM_TYPE_GRAPHS

    def frame_from_graph(self) -> jelly.RdfStreamFrame | None:
        """
        Emit current flow content (one graph) as jelly frame.

        Returns:
            jelly.RdfStreamFrame | None: jelly frame or none if
                flow is empty.

        """
        return self.to_stream_frame()


class DatasetsFrameFlow(FrameFlow):
    logical_type = jelly.LOGICAL_STREAM_TYPE_DATASETS

    def frame_from_dataset(self) -> jelly.RdfStreamFrame | None:
        """
        Emit current flow content (dataset) as jelly frame.

        Returns:
            jelly.RdfStreamFrame | None: jelly frame or none if
                flow is empty.

        """
        return self.to_stream_frame()


# TODO(Nastya): issue #184
FLOW_DISPATCH: dict[jelly.LogicalStreamType, type[FrameFlow]] = {
    jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES: FlatTriplesFrameFlow,
    jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS: FlatQuadsFrameFlow,
    jelly.LOGICAL_STREAM_TYPE_GRAPHS: GraphsFrameFlow,
    jelly.LOGICAL_STREAM_TYPE_DATASETS: DatasetsFrameFlow,
}


def flow_for_type(logical_type: jelly.LogicalStreamType) -> type[FrameFlow]:
    """
    Return flow based on logical type requested.

    Args:
        logical_type (jelly.LogicalStreamType): logical type requested.

    Raises:
        NotImplementedError: if logical type not supported.

    Returns:
        type[FrameFlow]: FrameFlow for respective logical type.

    """
    try:
        return FLOW_DISPATCH[logical_type]
    except KeyError:
        msg = (
            "unsupported logical stream type: "
            f"{jelly.LogicalStreamType.Name(logical_type)}"
        )
        raise NotImplementedError(msg) from None
